#include <moveit/move_group_interface/move_group_interface.h>
#include <moveit/planning_scene_interface/planning_scene_interface.h>
#include <moveit_msgs/DisplayRobotState.h>
#include <moveit_msgs/DisplayTrajectory.h>
#include <moveit_msgs/AttachedCollisionObject.h>
#include <moveit_msgs/CollisionObject.h>
#include <moveit_visual_tools/moveit_visual_tools.h>

#include <geometry_msgs/Pose.h>
#include <std_msgs/String.h>

#include <boost/foreach.hpp>
#include <boost/lexical_cast.hpp>

#include <remote_rosbag_record/call.hpp>
#include <boost/regex.hpp>

#include <tf2_ros/buffer.h>
#include <tf2_ros/transform_listener.h>
#include <tf2_geometry_msgs/tf2_geometry_msgs.h>

#include <boost/algorithm/string.hpp>
#include <iostream>
#include <fstream>
#include <string>

#include <typeinfo>

#define foreach BOOST_FOREACH

const double tau = 2 * M_PI;

class CSVReader{

  std::string fileName;
  std::string delimeter;

public:
  CSVReader(std::string filename, std::string delm = ",") :
          fileName(filename), delimeter(delm)
  { }

  //Function to fetch data
  std::vector<std::vector<std::string> > getData();
};


std::vector<std::vector<std::string> > CSVReader::getData(){

  std::ifstream file(fileName);

  std::vector<std::vector<std::string> >dataList;
  std::string line = "";

  //iterate through lines and split content
  while (getline(file, line)){

        std::vector<std::string> vec;
        boost::algorithm::split(vec, line, boost::is_any_of(delimeter));
        dataList.push_back(vec);

  }


  file.close();

  return dataList;

}

class xyAxes{

public: std::vector<double> values;
	
	xyAxes() {
		values = {0,0};
	}

	xyAxes(std::vector<double> input_values){
	
		values = input_values;
	}


};




int main(int argc, char** argv)
{
  ros::init(argc, argv, "eff-measurements");
  ros::NodeHandle node_handle;

  // ROS spinning must be running for the MoveGroupInterface to get information
  // about the robot's state. One way to do this is to start an AsyncSpinner
  // beforehand.
  ros::AsyncSpinner spinner(1);
  spinner.start();


  //
  // Setup
  // ^^^^^
  //
  // MoveIt operates on sets of joints called "planning groups" and stores them in an object called
  // the `JointModelGroup`. Throughout MoveIt the terms "planning group" and "joint model group"
  // are used interchangably.
  static const std::string PLANNING_GROUP = "manipulator";

  // The :planning_interface:`MoveGroupInterface` class can be easily
  // setup using just the name of the planning group you would like to control and plan for.
  moveit::planning_interface::MoveGroupInterface move_group_interface(PLANNING_GROUP);
  move_group_interface.setPlanningTime(10.0);

  // We will use the :planning_interface:`PlanningSceneInterface`
  // class to add and remove collision objects in our "virtual world" scene
  moveit::planning_interface::PlanningSceneInterface planning_scene_interface;

  // Raw pointers are frequently used to refer to the planning group for improved performance.
  const moveit::core::JointModelGroup* joint_model_group =
      move_group_interface.getCurrentState()->getJointModelGroup(PLANNING_GROUP);

  // Visualization
  // ^^^^^^^^^^^^^
  //
  // The package MoveItVisualTools provides many capabilities for visualizing objects, robots,
  // and trajectories in RViz as well as debugging tools such as step-by-step introspection of a script.
  namespace rvt = rviz_visual_tools;
  moveit_visual_tools::MoveItVisualTools visual_tools("base_link");
  visual_tools.deleteAllMarkers();

  // Remote control is an introspection tool that allows users to step through a high level script
  // via buttons and keyboard shortcuts in RViz
  visual_tools.loadRemoteControl();

  // RViz provides many types of markers, in this demo we will use text, cylinders, and spheres
  Eigen::Isometry3d text_pose = Eigen::Isometry3d::Identity();
  text_pose.translation().z() = 1.0;

  // Batch publishing is used to reduce the number of messages being sent to RViz for large visualizations
  visual_tools.trigger();

  // Getting Basic Information
  // ^^^^^^^^^^^^^^^^^^^^^^^^^
  //
  // We can print the name of the reference frame for this robot.
  ROS_INFO_NAMED("tutorial", "Planning frame: %s", move_group_interface.getPlanningFrame().c_str());

  // We can also print the name of the end-effector link for this group.
  ROS_INFO_NAMED("tutorial", "End effector link: %s", move_group_interface.getEndEffectorLink().c_str());

  // We can get a list of all the groups in the robot:
  ROS_INFO_NAMED("tutorial", "Available Planning Groups:");
  std::copy(move_group_interface.getJointModelGroupNames().begin(),
            move_group_interface.getJointModelGroupNames().end(), std::ostream_iterator<std::string>(std::cout, ", "));



  // Start
  // ^^^^^
  visual_tools.prompt("Press 'next' in the RvizVisualToolsGui window to start the demo, insert the scene");

  // Note that we are just planning, not asking move_group_interface
  // to actually move the robot.
  moveit::planning_interface::MoveGroupInterface::Plan my_plan;

  

  // Adding objects to the environment
  // ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
  //

  // Now let's define a collision object ROS message for the robot to avoid.
  
  moveit_msgs::CollisionObject co0;
  moveit_msgs::CollisionObject co1;
  moveit_msgs::CollisionObject co2;

  co0.header.frame_id = move_group_interface.getPlanningFrame();
  co1.header.frame_id = move_group_interface.getPlanningFrame();
  co2.header.frame_id = move_group_interface.getPlanningFrame();   
 
  // The id of the object is used to identify it
  co0.id = "box0";
  co1.id = "box1";  
  co2.id = "box2";

  // Define box0 to add to the world.
  shape_msgs::SolidPrimitive primitive;
  primitive.type = primitive.BOX;
  primitive.dimensions.resize(3);
  primitive.dimensions[primitive.BOX_X] = 1.76;
  primitive.dimensions[primitive.BOX_Y] = 0.7;
  primitive.dimensions[primitive.BOX_Z] = 0.03;
  // Define a pose for box0 (specified relative to frame_id)
  geometry_msgs::Pose box_pose;
  box_pose.orientation.w = 1.0;
  box_pose.position.x = -0.66;
  box_pose.position.y = 0.21;
  box_pose.position.z = -0.01;                                   
  co0.primitives.push_back(primitive);
  co0.primitive_poses.push_back(box_pose);
  co0.operation = co0.ADD;
                           

  // Define box1 to add to the world.
  primitive.dimensions[primitive.BOX_X] = 0.88;
  primitive.dimensions[primitive.BOX_Y] = 0.88;
  primitive.dimensions[primitive.BOX_Z] = 0.03;
  // Define a pose for box1 (specified relative to frame_id)
  box_pose.position.x = -1.1;
  box_pose.position.y = 1;
  box_pose.position.z = -0.01;                                         
  co1.primitives.push_back(primitive);
  co1.primitive_poses.push_back(box_pose);
  co1.operation = co1.ADD;


  // Define box2 to add to the world.
  primitive.dimensions[primitive.BOX_X] = 0.15;
  primitive.dimensions[primitive.BOX_Y] = 0.07;
  primitive.dimensions[primitive.BOX_Z] = 0.3;
  // Define a pose for box2 (specified relative to frame_id)
  box_pose.position.x = 0.0;
  box_pose.position.y = -0.18;
  box_pose.position.z = -0.15;
  co2.primitives.push_back(primitive);
  co2.primitive_poses.push_back(box_pose);
  co2.operation = co2.ADD;
  
  //all the collision objects
  std::vector<moveit_msgs::CollisionObject> collision_objects;
  collision_objects.push_back(co0);
  collision_objects.push_back(co1);
  collision_objects.push_back(co2);  


  // Now, let's add the collision object into the world
  // (using a vector that could contain additional objects)
  ROS_INFO_NAMED("tutorial", "Add objects into the world");
  planning_scene_interface.addCollisionObjects(collision_objects);
  //planning_scene_interface.applyCollisionObjects(collision_objects);



  // Read from csv file and store states
  // ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
  //
  //  *no header in csv, valus to be double
  //

  visual_tools.trigger();
  visual_tools.prompt("Press 'next' in the RvizVisualToolsGui window to read from csv");

  std::string traj_file = "sin";
  CSVReader reader("/home/ioana/catkin_ws/src/robot_identification/"+traj_file+".csv");

  std::vector<std::vector<std::string> > csvData = reader.getData();

  // create vector of xyaxis to store all xy states
  std::vector<xyAxes> all_xy;
  std::vector<double> auxVect(2);

  for (std::vector<std::string> vec:csvData){

        //changing strings from vec to double in auxVect          
        std::transform(vec.begin(), vec.end(), auxVect.begin(), [](const std::string& val)
        { return std::stod(val); }
        );

        all_xy.push_back(xyAxes(auxVect));

  }



  // check joints
  foreach( xyAxes xy, all_xy){

        std::cout << std::endl << "New state: " << std::endl;
        for (int i=0; i<2; i++ ){
                std::cout << xy.values[i] << ", ";
        }
  }



  // Cartesian Paths
  // ^^^^^^^^^^^^^^^
  
  visual_tools.trigger();
  visual_tools.prompt("Press 'next' in the RvizVisualToolsGui window to start planning for the end-effector");   
  

  //placeholders for the iteration poses
  geometry_msgs::PoseStamped target_pose;
  target_pose.header.frame_id = "astra_camera_link";
  //orientation to preset (perpendicular to door)
  target_pose.pose.orientation.x = 0.0;
  target_pose.pose.orientation.y = 0.0;
  target_pose.pose.orientation.z = 0.7071;
  target_pose.pose.orientation.w = 0.7071;

  std::vector<geometry_msgs::Pose> waypoints;
  
    
  //int lim = all_xy.size();

  foreach(xyAxes xy, all_xy ){
        // lim -= 1;
        // taking each x-th point
	// if ( lim%5 != 0 )
	//	continue;

         
	target_pose.pose.position.x = xy.values[0]/20;
	target_pose.pose.position.y = 0.35;
	target_pose.pose.position.z = xy.values[1]/20 + 0.8;
	
	waypoints.push_back(target_pose.pose);
   
  
  }
	
  
  // We want the Cartesian path to be interpolated at a resolution of 1 cm
  // which is why we will specify 0.01 as the max step in Cartesian
  // translation.  We will specify the jump threshold as 0.0, effectively disabling it.
  // Warning - disabling the jump threshold while operating real hardware can cause
  // large unpredictable motions of redundant joints and could be a safety issue

  moveit_msgs::RobotTrajectory trajectory;
  const double jump_threshold = 0.0;
  const double eef_step = 0.01;
  double fraction = move_group_interface.computeCartesianPath(waypoints, eef_step, jump_threshold, trajectory,true);
  ROS_INFO_NAMED("tutorial", "Visualizing plan (Cartesian path) (%.2f%% acheived)", fraction * 100.0);

  // Visualize the plan in RViz
  visual_tools.deleteAllMarkers();
  visual_tools.publishPath(waypoints, rvt::LIME_GREEN, rvt::SMALL);
  for (std::size_t i = 0; i < waypoints.size(); ++i)
    visual_tools.publishAxisLabeled(waypoints[i], std::to_string(i), rvt::SMALL);
  


  // Execution and Recording
  // ^^^^^^^^^^^^^^^^^^^^^^^

  // path of the folder the bag files will be saved in
  std::string path = "/home/ioana/catkin_ws/measurements_bag_files/";
  std::string filename = traj_file;
  // changing value in parameter server for the logging of the bag files
  node_handle.setParam("/robot_record/prefix",path+filename);



  visual_tools.trigger();             
  visual_tools.prompt("Press 'next' in the RvizVisualToolsGui window only if you want to go with the execution");

  remote_rosbag_record::call(boost::regex("/start"));
  ros::Duration(2).sleep();
  move_group_interface.execute(trajectory);
  ros::Duration(2).sleep();
  remote_rosbag_record::call(boost::regex("/stop"));

  
  visual_tools.trigger();
  visual_tools.prompt("Press 'next' to continue, removing objects");



  // Remove the objects from the world.
  // ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
  ROS_INFO_NAMED("tutorial", "Remove the objects from the world");
  std::vector<std::string> object_ids;
  object_ids.push_back(co0.id);
  object_ids.push_back(co1.id);
  object_ids.push_back(co2.id);
  planning_scene_interface.removeCollisionObjects(object_ids);

  visual_tools.trigger();
  visual_tools.prompt("Press 'next' in the RvizVisualToolsGui window to once the collision object disapears");


                                                
  
  // END_TUTORIAL


  ros::shutdown();
  return 0;
}

