/*********************************************************************
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2012, Willow Garage, Inc.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of Willow Garage nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *********************************************************************/

/* Author: Sachin Chitta, Michael Lautman */

#include <pluginlib/class_loader.h>
#include <ros/ros.h>

// MoveIt
#include <moveit/robot_model_loader/robot_model_loader.h>
#include <moveit/planning_interface/planning_interface.h>
#include <moveit/planning_scene/planning_scene.h>
#include <moveit/kinematic_constraints/utils.h>
#include <moveit_msgs/DisplayTrajectory.h>
#include <moveit_msgs/PlanningScene.h>
#include <moveit_visual_tools/moveit_visual_tools.h>
#include <moveit_msgs/RobotTrajectory.h>

#include <boost/scoped_ptr.hpp>
#include <boost/foreach.hpp>
#include <boost/lexical_cast.hpp>

#include <moveit/move_group_interface/move_group_interface.h>
#include <moveit/planning_scene_interface/planning_scene_interface.h>
#include <std_msgs/Float64.h>

#include <remote_rosbag_record/call.hpp>
#include <boost/regex.hpp>

#include <dirent.h>

#define foreach BOOST_FOREACH



class CSVReader{

  std::string fileName;
  std::string delimeter;

public:
  CSVReader(std::string filename, std::string delm = ",") :
          fileName(filename), delimeter(delm)
  { }

  //Function to fetch data
  std::vector<std::vector<std::string> > getData();
};


std::vector<std::vector<std::string> > CSVReader::getData(){

  std::ifstream file(fileName);

  std::vector<std::vector<std::string> >dataList;
  std::string line = "";

  //iterate through lines and split content
  while (getline(file, line)){

        std::vector<std::string> vec;
        boost::algorithm::split(vec, line, boost::is_any_of(delimeter));
        dataList.push_back(vec);

  }


  file.close();

  return dataList;

}

// class containing x(time) and y(amplitude) of a trajectory 
class xyAxes{
public: std::vector<double> values;

        xyAxes() {
                values = {0,0};
        }

        xyAxes(std::vector<double> input_values){

                values = input_values;
        }

};






int main(int argc, char** argv)
{
  const std::string node_name = "velocities_record";
  ros::init(argc, argv, node_name);
  ros::AsyncSpinner spinner(1);
  spinner.start();
  ros::NodeHandle node_handle("~");

  
  const std::string PLANNING_GROUP = "manipulator";
  robot_model_loader::RobotModelLoader robot_model_loader("robot_description");
  const moveit::core::RobotModelPtr& robot_model = robot_model_loader.getModel();
  /* Create a RobotState and JointModelGroup to keep track of the current robot pose and planning group*/
  moveit::core::RobotStatePtr robot_state(new moveit::core::RobotState(robot_model));
  const moveit::core::JointModelGroup* joint_model_group = robot_state->getJointModelGroup(PLANNING_GROUP);

  // Using the :moveit_core:`RobotModel`, we can construct a :planning_scene:`PlanningScene`
  // that maintains the state of the world (including the robot).
  planning_scene::PlanningScenePtr planning_scene(new planning_scene::PlanningScene(robot_model));

  // Configure a valid robot state
  planning_scene->getCurrentStateNonConst().setToDefaultValues(joint_model_group, "up");

  // We will now construct a loader to load a planner, by name.
  // Note that we are using the ROS pluginlib library here.
  boost::scoped_ptr<pluginlib::ClassLoader<planning_interface::PlannerManager>> planner_plugin_loader;
  planning_interface::PlannerManagerPtr planner_instance;
  std::string planner_plugin_name;

  // We will get the name of planning plugin we want to load
  // from the ROS parameter server, and then load the planner
  // making sure to catch all exceptions.
  if (!node_handle.getParam("planning_plugin", planner_plugin_name))
    ROS_FATAL_STREAM("Could not find planner plugin name");
  try
  {
    planner_plugin_loader.reset(new pluginlib::ClassLoader<planning_interface::PlannerManager>(
        "moveit_core", "planning_interface::PlannerManager"));
  }
  catch (pluginlib::PluginlibException& ex)
  {
    ROS_FATAL_STREAM("Exception while creating planning plugin loader " << ex.what());
  }
  try
  {
    planner_instance.reset(planner_plugin_loader->createUnmanagedInstance(planner_plugin_name));
    if (!planner_instance->initialize(robot_model, node_handle.getNamespace()))
      ROS_FATAL_STREAM("Could not initialize planner instance");
    ROS_INFO_STREAM("Using planning interface '" << planner_instance->getDescription() << "'");
  }
  catch (pluginlib::PluginlibException& ex)
  {
    const std::vector<std::string>& classes = planner_plugin_loader->getDeclaredClasses();
    std::stringstream ss;
    for (const auto& cls : classes)
      ss << cls << " ";
    ROS_ERROR_STREAM("Exception while loading planner '" << planner_plugin_name << "': " << ex.what() << std::endl
                                                         << "Available plugins: " << ss.str());
  }



  // Visualization
  // ^^^^^^^^^^^^^
  
  namespace rvt = rviz_visual_tools;
  moveit_visual_tools::MoveItVisualTools visual_tools("base_link");
  visual_tools.loadRobotStatePub("/display_robot_state");
  visual_tools.enableBatchPublishing();
  visual_tools.deleteAllMarkers(); 
  visual_tools.trigger();
  
  visual_tools.loadRemoteControl();

  Eigen::Isometry3d text_pose = Eigen::Isometry3d::Identity();
  text_pose.translation().z() = 1.75;
  visual_tools.publishText(text_pose, "Motion Planning API Demo", rvt::WHITE, rvt::XLARGE);

  visual_tools.trigger();
  visual_tools.prompt("Press 'next' in the RvizVisualToolsGui window to start the demo, after you insert the scene");
 
  visual_tools.publishRobotState(planning_scene->getCurrentStateNonConst(), rviz_visual_tools::GREEN);
  visual_tools.trigger();
  


  
  // Find the directory containing the csv files to iterate through
  // ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
 
  std::vector<std::string> files;
  struct dirent *entry;
  DIR *dir = opendir("/home/ioana/catkin_ws/src/robot_identification/csv_files/");

  while ((entry = readdir(dir)) !=NULL){
	files.push_back(entry->d_name);	
	if (files.back().find(".csv") == std::string::npos ){ 
            files.pop_back();}
  }
  closedir(dir);


  foreach (std::string csv_str, files){

      planning_scene->getCurrentStateNonConst().setToDefaultValues(joint_model_group, "up");	  
      std::cout<<"file found: "<<csv_str<<std::endl;
  
      planning_interface::MotionPlanRequest req;
      planning_interface::MotionPlanResponse res;
      geometry_msgs::PoseStamped pose;

      req.group_name = PLANNING_GROUP;
      planning_interface::PlanningContextPtr context;

      ros::Publisher display_publisher =   node_handle.advertise<moveit_msgs::DisplayTrajectory>("/move_group/display_planned_path", 1, true);

      moveit_msgs::DisplayTrajectory display_trajectory;
      moveit_msgs::MotionPlanResponse response;



      // Read csv file and store joint states
      // ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
      // 
      // *no header in csv, values to be double, in radians
      

      visual_tools.trigger();
      visual_tools.prompt("Press 'next' to in the RvizVisualToolsGui to read from the csv");
      std::string csv_path = "/home/ioana/catkin_ws/src/robot_identification/csv_files/";
      CSVReader reader(csv_path+csv_str);
      std::vector<std::vector<std::string> > csvData = reader.getData();

      //create vector of xyAxes to store all joint states
      std::vector<xyAxes> all_xy;
      std::vector<double> auxVect(2);

      for (std::vector<std::string> vec:csvData){
            //changing strings from vec to double in auxVect          
            std::transform(vec.begin(), vec.end(), auxVect.begin(), [](const std::string& val)
            { return std::stod(val); }
            );

            all_xy.push_back(xyAxes(auxVect));
      }

      // check joints
      foreach( xyAxes xy, all_xy){
          std::cout << std::endl << "New state: " << std::endl;
          for (int i=0; i<2; i++ ){
                  std::cout << xy.values[i] << ", ";
          }
      }




      // Joint Space Goals
      // ^^^^^^^^^^^^^^^^^
     
      visual_tools.trigger();
      visual_tools.prompt("Press 'next' in the RvizVisualToolsGui window to plan joint space");

      moveit::core::RobotState goal_state(robot_model);

      int count = 1;
      foreach (xyAxes xy, all_xy){
   
          std::vector<double> joint_sin(6,0.0);	  
          // choosing the joint to move
	  joint_sin[3] = xy.values[1];
    
          double offset = -1.5707;
          joint_sin[1] += offset;
          joint_sin[3] += offset;
    
          std::vector<double> joint_values(6);
          joint_values = joint_sin;

          goal_state.setJointGroupPositions(joint_model_group, joint_values);
          moveit_msgs::Constraints joint_goal = kinematic_constraints::constructGoalConstraints(goal_state, joint_model_group);
  
          req.goal_constraints.clear();
          req.goal_constraints.push_back(joint_goal);
   
          // Call the planner and visualize the trajectory
          // Re-construct the planning context 
          context = planner_instance->getPlanningContext(planning_scene, req, res.error_code_);
          // Call the Planner 
          context->solve(res);
          // Check that the planning was successful 
          if (res.error_code_.val != res.error_code_.SUCCESS)
          {
            ROS_ERROR("Could not compute plan successfully");
            return 0;
          }

          // Visualize the trajectory 
          res.getMessage(response);
          display_trajectory.trajectory.push_back(response.trajectory);

          // Now you should see two planned trajectories in series
          visual_tools.publishTrajectoryLine(display_trajectory.trajectory.back(), joint_model_group);
          visual_tools.trigger();
          display_publisher.publish(display_trajectory);

          // Set the state in the planning scene to the final state of the last plan 
          robot_state->setJointGroupPositions(joint_model_group, response.trajectory.joint_trajectory.points.back().positions);
          planning_scene->setCurrentState(*robot_state.get());
          std::cout << count << " of "<< all_xy.size() << std::endl;

          count += 1;

          // Display the goal state
          visual_tools.publishRobotState(planning_scene->getCurrentStateNonConst(), rviz_visual_tools::GREEN);
          visual_tools.publishAxisLabeled(pose.pose, "goal");
          visual_tools.publishText(text_pose, "Joint Space Goal", rvt::WHITE, rvt::XLARGE);
  
      }

  


      // EXECUTION
      // ^^^^^^^^^
      //
  
      //time has to be added
  
      int disp_traj_nr = display_trajectory.trajectory.size();
      int sz_tr;

      std::cout<<"nr of trajectories: "<<disp_traj_nr<<std::endl;

      double time_dur= 0.0;
  

      for (int t=0;t<disp_traj_nr;t++){
  
        sz_tr = display_trajectory.trajectory[t].joint_trajectory.points.size();
        std::cout<<"for trajectory "<<t<<" there are "<<sz_tr<<" points"<<std::endl;

        for (int i=0;i<sz_tr;i++){
    
          ros::Duration time_dur_ros = ros::Duration(time_dur);
          display_trajectory.trajectory[t].joint_trajectory.points[i].time_from_start = time_dur_ros;
          time_dur += all_xy[1].values[0];
	  std::cout<<"time duration: "<<time_dur<<std::endl;
         }
      }


      // changing the name of the bag file by changing value in parametere server
      std::string path = "/home/ioana/catkin_ws/measurements_bag_files/ramp_velocities/";
      std::string filename = csv_str.erase(csv_str.size()-4);
      node_handle.setParam("/robot_record/prefix",path+filename);


      //actual execution
      visual_tools.trigger();
      visual_tools.prompt("Press 'next' in the RvizVisualToolsGui window only if you want to go with the execution");

      moveit::planning_interface::MoveGroupInterface move_group_interface(PLANNING_GROUP);
      move_group_interface.setPlanningTime(10.0);
      moveit::planning_interface::MoveGroupInterface::Plan my_plan;

 
      // adding header and joint names in the plan struct
      my_plan.trajectory_.joint_trajectory.header=display_trajectory.trajectory[0].joint_trajectory.header;
      my_plan.trajectory_.joint_trajectory.joint_names=display_trajectory.trajectory[0].joint_trajectory.joint_names;


      // adding points to the plan joint trajectory points vector  
      foreach (moveit_msgs::RobotTrajectory traj_part, display_trajectory.trajectory){
	  int nr_points = traj_part.joint_trajectory.points.size();

	  for (int i=0;i<nr_points;i++){	
		my_plan.trajectory_.joint_trajectory.points.push_back(traj_part.joint_trajectory.points[i]);
	  }
      }
	   
      remote_rosbag_record::call(boost::regex("/start"));
      move_group_interface.execute(my_plan);
      remote_rosbag_record::call(boost::regex("/stop"));
      //END of one movement

      visual_tools.trigger();
      visual_tools.prompt("Press 'next' in the RvizVisualToolsGui window to go to next measurement");
      visual_tools.deleteAllMarkers(); 

      //resetting the planner
      try
      {
         planner_instance.reset(planner_plugin_loader->createUnmanagedInstance(planner_plugin_name));
         if (!planner_instance->initialize(robot_model, node_handle.getNamespace()))
           ROS_FATAL_STREAM("Could not initialize planner instance");
           ROS_INFO_STREAM("Using planning interface '" << planner_instance->getDescription() << "'");
       }
      catch (pluginlib::PluginlibException& ex)
      {
         const std::vector<std::string>& classes = planner_plugin_loader->getDeclaredClasses();
         std::stringstream ss;
         for (const auto& cls : classes)
           ss << cls << " ";
         ROS_ERROR_STREAM("Exception while loading planner '" << planner_plugin_name << "': " << ex.what() << std::endl
                                                         << "Available plugins: " << ss.str());
       }
  
  
  

  }

  visual_tools.trigger();
  visual_tools.prompt("End of the node");
  planner_instance.reset();

  return 0;
}
